//------------------------------------------------------------------------------
// firmware для ATtiny85, используемой как АЦП с интерфейсом 1-Wire (slave)
// Протокол упрощен (только одно устройство на линии и отсутствие CRC)
// OldBean, 13.10.2018
//------------------------------------------------------------------------------
#include <avr/io.h>
#include <avr/interrupt.h>

#define F_CPU 16000000UL
#include <util/delay.h>

//------------------------------------------------------------------------------
static uint8_t ocr0A = 255; // Значение регистра сравнения A таймера 0
static uint8_t flagADC = 0; // Флаг разрешения запуска АЦП
static uint8_t cadc = 0, cADCw; // Вспомогательные переменные
static uint8_t wrk = 0;

static uint8_t wFlag = 0; // Состояние записи в линию 1-Wire
static uint8_t srFlag = 0; // Флаг состояния "skip ROM" (игнорируется адрес)
static uint16_t tau; // Длительность низкого уровня после старта тайм-слота
static uint8_t mask8 = 0x01; // Битовая маска для байта
static uint16_t mask16 = 0x0001; // Битовая маска для двухбайтового слова
static uint8_t data = 0;

static uint16_t adcv; // Код АЦП, передаваемый мастеру шины
static uint32_t sum; // Сумматор для усреднения
static uint16_t cntr; // Счетчик просуммированных выборок
volatile static uint8_t adcFlag = 0; // Если 1, то измерения разрешены
//------------------------------------------------------------------------------
ISR(PCINT0_vect){ // Обработчик внешнего прерывания на PB3
  if(!(PINB & _BV(PB3))) { // Был переход линии с высокого уровня на низкий
    GTCCR = _BV(PSR0); // Сброс предделителя таймера 0
    TCNT0 = 0; // Обнуляем счетчик таймера 0
    if(wFlag) { // Пишем бит в линию (а байты - сначала младший затем старший)
      if(!(adcv & mask16)) { // Нужно писать 0
        PORTB &= ~_BV(PB3); // Устанавливаем низкий уровень на пине PB3
        DDRB |= _BV(PB3); // Переводим PB3 в режим записи
        _delay_us(20); // Держим паузу (15 мкс с небольшим запасом ;)
        DDRB &= ~_BV(PB3); // Отпускаем линию (PB3 - в режим чтения)
      } // А 1 получится автоматически, если ничего не делать ;)
      mask16 <<= 1; // Модифицируем маску (переход на следующий бит)
    }
  } else { // Наоборот - переход линии с низкого на высокий
    // Оценим длительность импульса (низкого уровня на линии) tau в мкс
    tau = TCNT0; tau <<= 2; // Единица счетчика TIMER0 - это 4 мкс
    if(tau >= 480) { // Это RESET. Для простоты импульс PRESENCE пока не подаем
      wFlag = 0; // Переходим в режим чтения
      srFlag = 0; // Сбрасываем состояние "skip ROM"
      mask8 = 0x01; // Сбрасываем маску на младший бит
      data = 0; // Очищаем принимаемый байт
    } else { // Слот или кусок слота
      if(!wFlag) { // Режим чтения
        if(tau < 15) { // Получен единичный бит
          data |= mask8;
        }
        mask8 <<= 1; // Модифицируем маску
        if(mask8 == 0) { // Считали все восемь бит (маска полностью ушла влево)
          switch(data) { // Анализируем полученную команду
            case 0xcc: // Получена команда игнорирования адреса (skip ROM)
              srFlag = 1;
              break;
            case 0x44: // Получена команда запуск преобразования
              sum = 0; // Очищаем сумматор
              cntr = 0; // Сброс счетчика выборок
              adcFlag = 1; // Разрешаем измерение и суммирование
              break;
            case 0xbe: // Мастер перешел в режим чтения, а slave - записи
              if(srFlag) { // Работаем без адреса (была команда skip ROM)
                wFlag = 1; // Устанавливаем флаг записи на линю
              } // В противном случае (без skip ROM) - игнорируем
              break;
          }
          mask8 = 0x01; // Сбрасываем маску на младший бит
          mask16 = 0x0001;
          data = 0; // Очищаем принимаемый байт
        }
      } // Режим чтения
    }
  }
}
//------------------------------------------------------------------------------
int main(void) {
  DDRB |= _BV(PB0); // Пин для диагностики
  // Инициализация внешних прерываний на пине PB3 ------------------------------
  GIMSK = _BV(PCIE);   // Прерывания по изменению уровня на пине
  PCMSK = _BV(PCINT3); // Внешние прерывания ловим на пине PB3
  // Инициализация таймера 0 для тактирования АЦП ------------------------------
  // Счетчик работает в нормальном режиме
  TCCR0B = _BV(CS01) | _BV(CS00); // Предделитель на 64
  // Инициализация АЦП ---------------------------------------------------------
  ADMUX = 0x00; // В качестве источника опорного напряжения - питание
  ADMUX |= _BV(MUX1); // Работаем в ADC2 (источник подключен к пину PB4)
  ADCSRA = _BV(ADEN); // АЦП включен
  ADCSRA |= _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0); // Предделитель на 128
  sei(); // Разрешаем глобальные прерывания
  int flag = 0;
  while(1) { // Бесконечный цикл
    if(!adcFlag) continue; // Измерения запрещены
    ADCSRA |= _BV(ADSC); // Запуск АЦП
    while(ADCSRA & _BV(ADSC)); // Ждем пока завершится преобразование

    // Часть операций ниже не являются атомарными, но в периоды времени,
    // когда они выполняются никаких прерываний не инициируется, т.к.
    // мастер делает паузу примерно на 750 мс и не "дергает" шину.
    // В связи с этим никаких дополнительных мер по безопасности следующего
    // ниже кода не применяется
    sum += ADCL;
    sum += (ADCH << 8);
    if(++cntr == 4096) { // Достаточно насуммировали...
      sum += 2048; // Ну усреднять, так усреднять ;) 
      adcv = sum >> 12; // Усредненный по 4096 выборкам код АЦП
      adcFlag = 0; // Дальнейшие измерения до следующей команды 0x44 запрещаем
    }
  }
  return 0;
}
